"""Comp Neuro Exception Classes

.. note::

    All anticipated exceptions should raise custom Exceptions so that user
    can identify when an error is not anticipated and help with debugging.
"""


class CompNeuroError(Exception):
    """Base Comp Neuro Exceptions"""


class CompNeuroUtilsError(CompNeuroError):
    """Base Comp Neuro Utility Exceptions"""


class CompNeuroPlotError(CompNeuroUtilsError):
    """Comp Neuro Plotter Exceptions"""


class CompNeuroSignalError(CompNeuroUtilsError):
    """Comp Neuro Signal Exceptions"""


class CompNeuroModelError(CompNeuroError):
    """Comp Neuro Model Exceptions"""



import numpy as np


def hh(t, I_ext):
    # Assume that the time is given in seconds, and convert it to
    # number of milliseconds
    t = 1000 * t
    dt = t[1] - t[0]

    # Conductances and reverse potentials for K, Na, R:
    E = [-12, 115, 10.613]
    g = [36, 120, 0.300]
    x = [0, 0, 1.000]

    # Initialize membrane voltage:
    V = np.zeros(len(t))
    V[0] = -10

    # Initialize currents due to the various ions:
    I = np.zeros((len(t), 3))

    # Alpha and beta variables:
    a = np.zeros(3)
    b = np.zeros(3)

    # Channel activations:
    gnmh = np.zeros((len(t), 3))

    # Perform numerical integration of the ODEs:
    for i in range(1, len(t)):

        a[0] = (10 - V[i - 1]) / (100 * (np.exp((10 - V[i - 1]) / 10) - 1))
        a[1] = (25 - V[i - 1]) / (10 * (np.exp((25 - V[i - 1]) / 10) - 1))
        a[2] = 0.07 * np.exp(-V[i - 1] / 20)

        b[0] = 0.125 * np.exp(-V[i - 1] / 80)
        b[1] = 4 * np.exp(-V[i - 1] / 18)
        b[2] = 1 / (np.exp((30 - V[i - 1]) / 10) + 1)

        tau = 1 / (a + b)
        x0 = a * tau

        x = (1 - dt / tau) * x + dt / tau * x0

        gnmh[i, 0] = g[0] * x[0] ** 4
        gnmh[i, 1] = g[1] * x[1] ** 3 * x[2]
        gnmh[i, 2] = g[2]

        # Update the ionic currents and membrane voltage:
        I[i] = gnmh[i] * (V[i - 1] - E)
        V[i] = V[i - 1] + dt * (I_ext[i] - np.sum(I[i]))

    return V, I


def hhr(t, I_ext):
    # Assume that the time is given in seconds, and convert it to
    # number of milliseconds
    t = 1000 * t
    dt = t[1] - t[0]

    # Conductances and reverse potentials for K, Na, R:
    E = [-12, 115, 10.613]
    g = [36, 120, 0.300]

    # Normalize capacitance to 1 microfarad:
    C = 1

    # Initialize membrane voltage:
    V = np.zeros(len(t))
    V[0] = -10

    # Initialize currents due to the various ions:
    I = np.zeros((len(t), 3))

    # Initialize W:
    W = np.zeros(len(t))

    # Determine the slope S of the Rinzel approximation (see [1]):
    h_0 = 0.07 / (0.07 + 1 / (np.exp(3) + 1))
    n_0 = 0.1 / (np.exp(1) - 1) / (0.1 / (np.exp(1) - 1) + 0.125)
    S = (1 - h_0) / n_0

    # Perform numerical integration of the ODEs:
    for i in range(1, len(t)):

        alpha_n = 0.01 * (10 - V[i - 1]) / (np.exp((10 - V[i - 1]) / 10) - 1)
        alpha_m = 0.1 * (25 - V[i - 1]) / (np.exp((25 - V[i - 1]) / 10) - 1)
        alpha_h = 0.07 * np.exp(-V[i - 1] / 20)

        beta_n = 0.125 * np.exp(-V[i - 1] / 80)
        beta_m = 4 * np.exp(-V[i - 1] / 18)
        beta_h = 1 / (np.exp((30 - V[i - 1]) / 10) + 1)

        n_infty = alpha_n / (alpha_n + beta_n)
        m_infty = alpha_m / (alpha_m + beta_m)
        h_infty = alpha_h / (alpha_h + beta_h)
        W_infty = S / (1 + S ** 2) * (n_infty + S * (1 - h_infty))

        tau_W = 1 + 5 * np.exp(-((V[i - 1] - 10) ** 2) / 55 ** 2)

        # Use exponential Euler numerical integration method to compute W:
        # Recall that for exponential Euler, if dy/dt=a-b*y,
        # y(t_k+1)=y_tk*D + a/b*(1-D), where D=exp(-b*dt)
        a = 3 * W_infty / tau_W
        b = 3 / tau_W
        D = np.exp(-b * dt)
        W[i] = W[i - 1] * D + (a / b) * (1 - D)

        # Update the ionic currents and membrane voltage:
        I[i, 0] = g[0] * (W[i - 1] / S) ** 4 * (V[i - 1] - E[0])
        I[i, 1] = g[1] * m_infty ** 3 * (1 - W[i - 1]) * (V[i - 1] - E[1])
        I[i, 2] = g[2] * (V[i - 1] - E[2])
        V[i] = V[i - 1] + dt / C * (I_ext[i] - np.sum(I[i]))

    return V, W


def hhw(t, I_ext):
    # Assume that the time is given in seconds, and convert it to
    # number of milliseconds
    t = 1000 * t
    dt = t[1] - t[0]
    # Conductances and reverse potentials for Na, K
    E = [55.0, -92.0]
    g = [1.0, 26.0]
    # Time constant for R
    tau_R = 1.9
    # Normalize capacitance to 1 microfarad:
    C = 1.2
    # Initialize membrane voltage:
    V = np.zeros(len(t))
    V[0] = -70.0
    # Initialize currents due to the various ions:
    I = np.zeros((len(t), 2))
    # Initialize W:
    R = np.zeros(len(t))
    R[0] = 0.088

    # Perform numerical integration of the ODEs:
    for i in range(1, len(t)):
        R_infty = 0.0135 * V[i - 1] + 1.03

        # Use exponential Euler numerical integration method to compute W:
        # Recall that for exponential Euler, if dy/dt=a-b*y,
        # y(t_k+1)=y_tk*D + a/b*(1-D), where D=exp(-b*dt)
        a = R_infty / tau_R
        b = 1 / tau_R
        D = np.exp(-b * dt)
        R[i] = R[i - 1] * D + (a / b) * (1 - D)

        # Update the ionic currents and membrane voltage:
        I[i] = (
            g
            * np.array(
                [(17.81 + 0.4771 * V[i - 1] + 0.003263 * V[i - 1] ** 2), R[i - 1]]
            )
            * (V[i - 1] - E)
        )
        V[i] = V[i - 1] + dt / C * (I_ext[i] - np.sum(I[i]))

    return V, R



"""Test Utility Module of CompNeuro

Tests:
    1. Plot Submodule
    2. Signal Submodule
"""
import pytest
import numpy as np
import matplotlib.pyplot as plt
from compneuro import plot, signal
import compneuro.errors as err

DT = 1e-4
DUR = 1.0
T = np.arange(0, DUR, DT)


@pytest.fixture
def signal_data():
    dt = 1e-4
    dur = 1.0
    t = np.arange(0, 1.0, dt)
    bw = 100  # 30 Hz
    num = 2
    seed = 0
    return dt, dur, t, bw, num, seed


@pytest.fixture
def signal_spikes(signal_data):
    dt, dur, t, bw, num, seed = signal_data

    rate = 100
    randvar = np.random.rand(len(t), num)
    spikes = randvar < rate * dt
    return rate, spikes


@pytest.fixture
def matrix_data():
    return np.random.rand(100, len(T))


@pytest.fixture
def spike_data():
    spikes = np.random.rand(100, len(T)) < 0.5
    return spikes.astype(float)


def test_plot_mat(matrix_data):
    ax, cbar = plot.plot_mat(matrix_data, dt=DT, cax=True)
    assert isinstance(ax, plt.Axes)
    (ax,) = plot.plot_mat(matrix_data, dt=DT, cax=False)
    assert isinstance(ax, plt.Axes)
    (ax,) = plot.plot_mat(matrix_data, t=T, cax=False)
    assert isinstance(ax, plt.Axes)

    fig, ax = plt.subplots(1, 1)
    (ax2,) = plot.plot_mat(matrix_data, t=T, cax=False, ax=ax)
    assert ax == ax2

    fig, axes = plt.subplots(1, 2)
    ax, cbar = plot.plot_mat(matrix_data, t=T, ax=axes[0], cax=axes[1])
    assert ax == axes[0]

    fig, axes = plt.subplots(1, 2)
    ax, cbar = plot.plot_mat(
        matrix_data,
        t=T,
        y=np.arange(matrix_data.shape[0]),
        ax=axes[0],
        cax=axes[1],
        cbar_kw={"orientation": "horizontal"},
    )
    assert ax == axes[0]

    with pytest.raises(
        err.CompNeuroPlotError, match=r"Time vector .* does not have the same shape"
    ):
        ax, cbar = plot.plot_mat(matrix_data, t=[0], cax=True)

    with pytest.raises(
        err.CompNeuroPlotError, match=r"Time vector .* does not have the same shape"
    ):
        ax, cbar = plot.plot_mat(matrix_data.T, t=T, cax=True)


def test_plot_spikes(spike_data):
    ax = plot.plot_spikes(spike_data)
    assert isinstance(ax, plt.Axes)

    ax = plot.plot_spikes(spike_data, t=T)
    assert isinstance(ax, plt.Axes)

    ax = plot.plot_spikes(spike_data, dt=DT)
    assert isinstance(ax, plt.Axes)

    with pytest.raises(
        err.CompNeuroPlotError, match=r"Time vector .* does not have the same shape"
    ):
        ax = plot.plot_spikes(spike_data.T, t=[0])

    with pytest.raises(err.CompNeuroPlotError, match=r"dt must be a scalar value"):
        ax = plot.plot_spikes(spike_data.T, dt=[DT])

    with pytest.raises(err.CompNeuroPlotError, match=r"dt must be a scalar value"):
        ax = plot.plot_spikes(spike_data.T, dt=T)


def test_generate_stimulus(signal_data):
    dt, dur, t, bw, num, seed = signal_data
    step = signal.generate_stimulus("step", dt, dur, (0.2 * dur, 0.8 * dur), 100)
    para = signal.generate_stimulus("parabola", dt, dur, (0.2 * dur, 0.8 * dur), 100)
    ramp = signal.generate_stimulus("ramp", dt, dur, (0.2 * dur, 0.8 * dur), 100)

    assert len(step) == len(t)
    assert len(para) == len(t)
    assert len(ramp) == len(t)
    assert step.max() == 100
    assert para.max() == 100
    assert ramp.max() == 100

    amps = np.linspace(1, 100, num)
    step = signal.generate_stimulus("step", dt, dur, (0.2 * dur, 0.8 * dur), amps)
    para = signal.generate_stimulus("parabola", dt, dur, (0.2 * dur, 0.8 * dur), amps)
    ramp = signal.generate_stimulus("ramp", dt, dur, (0.2 * dur, 0.8 * dur), amps)

    assert step.shape == (len(amps), len(t))
    assert para.shape == (len(amps), len(t))
    assert ramp.shape == (len(amps), len(t))

    np.testing.assert_equal(step.max(axis=1), amps)
    np.testing.assert_equal(para.max(axis=1), amps)
    np.testing.assert_equal(ramp.max(axis=1), amps)


def test_generate_spike_from_psth(signal_data):
    dt, dur, t, bw, num, seed = signal_data
    rate = 100
    step = signal.generate_stimulus("step", dt, dur, (0.2 * dur, 0.8 * dur), rate)
    ss = signal.generate_spike_from_psth(dt, step, num=num, seed=seed)
    assert ss.shape == (num, len(t))
    assert np.sum(ss[:, np.logical_and(t < 0.2 * dur, t > 0.8 * dur)]) == 0
    assert np.max(np.abs(np.sum(ss, axis=1) / (0.6 * dur) - rate) / rate) < 0.2

    ss = signal.generate_spike_from_psth(dt, step, num=1, seed=seed)
    assert ss.shape == (len(t),)


def test_compute_psth(signal_data, signal_spikes):
    dt, dur, t, bw, num, seed = signal_data
    rate, ss = signal_spikes
    psth, psth_t = signal.compute_psth(ss, dt, window=2e-2, interval=1e-2)
    assert np.abs(np.mean(psth) - rate) / rate < 0.2


def test_snr(signal_data):
    dt, dur, t, bw, num, seed = signal_data

    amps = np.arange(0, 100, num)
    step = signal.generate_stimulus("step", dt, dur, (0.2 * dur, 0.8 * dur), amps)
    snr_inf = signal.snr(step, step)
    assert snr_inf.shape == step.shape
    assert np.all(snr_inf == np.inf)


def test_random_signal(signal_data):
    dt, dur, t, bw, num, seed = signal_data
    sig = signal.random_signal(t, bw, num, seed=seed)
    assert sig.shape == (num, len(t))

    # test Power
    for v in np.mean(sig ** 2, axis=-1):
        assert np.abs(v - 1) < 1e-10

    # test Bandwidth

    # test seed
    sig2 = signal.random_signal(t, bw, num, seed=seed)
    np.testing.assert_equal(sig, sig2)

    # test RNG
    rng = np.random.RandomState(seed)
    sig1_1 = signal.random_signal(t, bw, num, seed=rng)
    sig1_2 = signal.random_signal(t, bw, num, seed=rng)
    sig1_3 = signal.random_signal(t, bw, num, seed=rng)
    rng = np.random.RandomState(seed)
    sig2_1 = signal.random_signal(t, bw, num, seed=rng)
    sig2_2 = signal.random_signal(t, bw, num, seed=rng)
    sig2_3 = signal.random_signal(t, bw, num, seed=rng)
    np.testing.assert_equal(sig1_1, sig2_1)
    np.testing.assert_equal(sig1_2, sig2_2)
    np.testing.assert_equal(sig1_3, sig2_3)


def test_spikes_detect():
    volt = np.array([0.0, 1.0, 0.0, 0.0, 5.0, 2.0, 0.0])
    mask_ref = np.array([False, True, False, False, True, False, False])
    mask = signal.spike_detect(volt)
    np.testing.assert_equal(mask_ref, mask)

    volt = np.array([0.0, 1.0, 0.0, 0.0, 5.0, 2.0, 0.0])
    mask_ref = np.array([False, False, False, False, True, False, False])
    mask = signal.spike_detect(volt, threshold=2.0)
    np.testing.assert_equal(mask_ref, mask)

    volt = np.array(
        [[0.0, 1.0, 0.0, 0.0, 5.0, 2.0, 0.0], [0.0, 0.0, 5.0, 0.0, 0.0, 2.0, 0.0]]
    )
    mask_ref = np.array(
        [
            [False, True, False, False, True, False, False],
            [False, False, True, False, False, True, False],
        ]
    )
    mask = signal.spike_detect(volt)
    np.testing.assert_equal(mask_ref, mask)



# pylint:disable=C0103
"""Base Class for ODE-based Models
"""
import numpy as np
from numpy.core.numeric import full
from tqdm.auto import tqdm  # pylint:disable=F0002
from warnings import warn
from collections import OrderedDict
from numpy.core.fromnumeric import var
from scipy.integrate import solve_ivp, odeint
from scipy.integrate._ivp.base import OdeSolver
from scipy.interpolate import interp1d
from . import errors as err
from . import types as tpe


class BaseModel:
    """Base Model Class

    This base class is an opinionated implementation of ODE-based models
    for Comp Neuro.

    Arguments:
        num: number of state variables to simulate simultaneously
        kwargs: keyword arguments that overwrite initial conditions of state
            variables and values of parameters Can either be a `scalar` or
            a 1D numpy array with the dimensionality of :code:`(num,)`
    """

    Time_Scale: float = 1.0
    """Scaling for the time-step :code:`dt`. For models that must be simulated
    on a milisecond scale, set to :code:`1000`."""

    Default_Params: OrderedDict = OrderedDict()
    """A dictionary of (name, value) pair for parameters"""

    Default_States: OrderedDict = OrderedDict()
    """A dictionary of state variables where each entry is either

    - scalar: initial value of state variable
    - 3 tuple: (initial value, min, max) of the state variable

    .. warning::

        Due to a lack of variable bounding API in scipy, bounding variables
        are currently only available for :code:`Euler` method when solving
        ODEs.

        .. seealso:: :py:func:`solve`
    """

    def __init__(self, num: int = 1, **kwargs):
        if not isinstance(self.Default_States, OrderedDict):
            warn(
                "States need to be OrderedDict to maintain variable ordering, "
                "converting..."
            )
            self.Default_States = OrderedDict(self.Default_States)

        duplicate_vars = set(self.Default_Params.keys()).intersection(
            set(self.Default_States.keys())
        )
        if len(duplicate_vars) > 0:
            raise err.CompNeuroModelError(
                f"Params and States cannot duplicate names: {duplicate_vars}"
            )
        self.num = num
        self.params = OrderedDict(self.Default_Params.copy())
        self.states = OrderedDict()
        self.initial_states = OrderedDict()
        self.bounds = OrderedDict()

        for var_name, var_val in self.Default_States.items():
            var_val = np.atleast_1d(var_val)
            if len(var_val) == 1:
                self.initial_states[var_name] = var_val[[0]].copy()
                self.states[var_name] = var_val[[0]].copy()
                self.bounds[var_name] = None
            elif len(var_val) == 3:
                self.initial_states[var_name] = var_val[[0]].copy()
                self.states[var_name] = var_val[[0]].copy()
                self.bounds[var_name] = var_val[1:].copy()
            else:
                raise err.CompNeuroModelError(
                    f"Expect state variable {var_name} to have length 1 or 3, "
                    f"got {len(var_val)}"
                )

        for key, val in kwargs.items():
            if key in self.params:
                self.params[key] = val
            elif key in self.states:
                self.states[key] = val
                self.initial_states[key] = val
            else:
                raise err.CompNeuroModelError(f"Unrecognized argument {key}")
        self.__check_dimensions()

    def reset(self) -> None:
        """Reset Initial Values of the System"""
        for var_name, var_val in self.initial_states.items():
            var_val = np.atleast_1d(var_val)
            self.states[var_name] = var_val.copy()

    @property
    def _state_vec(self) -> np.ndarray:
        return np.vstack(list(self.states.values()))

    @_state_vec.setter
    def _state_vec(self, new_value) -> None:
        for var_name, new_val in zip(self.states.keys(), new_value):
            self.states[var_name] = new_val

    def __check_dimensions(self):
        for key, val in self.params.items():
            if np.isscalar(val):
                continue
            else:
                if len(val) == 1:
                    self.params[key] = np.repeat(val, self.num)
                else:
                    if len(val) != self.num:
                        raise err.CompNeuroModelError(
                            f"Parameter '{key}'' should have scalar value or array of "
                            f"length num={self.num}, got {len(val)} instead."
                        )
        for key, val in self.states.items():
            if np.isscalar(val):
                self.states[key] = np.full((self.num,), val)
            else:
                if len(val) == 1:
                    self.states[key] = np.repeat(val, self.num)
                else:
                    if len(val) != self.num:
                        raise err.CompNeuroModelError(
                            f"State '{key}' should have scalar value or array of "
                            f"length num={self.num}, got {len(val)} instead."
                        )

        for key, val in self.initial_states.items():
            if np.isscalar(val):
                self.initial_states[key] = np.full((self.num,), val)
            else:
                if len(val) == 1:
                    self.initial_states[key] = np.repeat(val, self.num)
                else:
                    if len(val) != self.num:
                        raise err.CompNeuroModelError(
                            f"Initial State '{key}' should have scalar value or array of "
                            f"length num={self.num}, got {len(val)} instead."
                        )

    def ode(self, t: float, states: np.ndarray, I_ext: float = None) -> np.ndarray:
        """Definition of Differential Equation

        .. note::

            For dynamical models described by ODEs, the function
            definition of the model should be of the form
            :math:`dx/dt=f(t,\mathbf{x})`,
            where :math:`\mathbf{x}` is usually a vector variable that describes
            the current state values at which the gradient is evaluted, and
            :math:`t` is the current time. Also note that :code:`t` is required
            to be compatible with scipy ode solver's API, although it is not
            needed for `autonomous` ODEs.

        .. seealso:: :py:mod:`scipy.integrate.solve_ivp`
            and :py:mod:`scipy.integrate.odeint`

        Arguments:
            t: current time value
            states: vector of state variable
            I_ext: external input current, optional

        Returns:
            A vector of gradient value evaluated based on the
            model dynamics. Should have the same shape as the
            states input vector.
        """

    def _ode_update(self, t: float, y: np.ndarray, *args, **kwargs) -> np.ndarray:
        """The actual update function that is passed to ode solvers"""

    def solve(
        self,
        t: np.ndarray,
        I_ext: np.ndarray = None,
        solver: tpe.solvers = "Euler",
        reset_initial_state: bool = True,
        verbose: bool = True,
        full_output: bool = False,
        **solver_kws,
    ) -> np.ndarray:
        """Solve model equation for entire input

        Arguments:
            t: 1d numpy array of time vector of the simulation
            I_ext: 1d/2d numpy array of external input current of shape

                - 1D: :code:`(len(t),)`, where the same input is shared for all model instances.
                - 2D: :code:`(len(t), num)` where :code:`num` is the number of components to simulate. The value `num` must be equal to the `self.num` value.

            solver: Which ODE solver to use, defaults to `Euler` which uses
                custom forward euler solver. Optional methods to use are:

                - `Euler`: Custom forward euler solver, default.
                - `odeint`: Use :py:mod:`scipy.integrate.odeint` which uses LSODA
                - `LSODA`: Use :py:mod:`scipy.integrate.odeint` which uses LSODA
                - `RK45/RK23/DOP853/Radau/BDF`: Use :py:mod:`scipy.integrate.solve_ivp` with the specified method
                - :py:mod:`scipy.integrate.OdeSolver` instance: Use :py:mod:`scipy.integrate.solve_ivp` with the provided custom solver

            reset_initial_state: whether to reset the initial state value of the
                model to the values in :code:`Default_State`. Default to True.
            verbose: whether to show progress
            full_output: whether to return the entire output from scipy's
                    ode solvers.

        .. note::

            String names for :code:`solve_ivp` (RK45/RK23/DOP853/Radau/BDF)
            are case-sensitive but not for any other methods.


        Keyword Arguments:
            solver_kws: arguments to be passed into the ode solvers if scipy
                solvers are used.

                .. seealso: :py:mod:`scipy.integrate.solve_ivp` and
                    :py:mod:`scipy.integrate.odeint`

        Returns:
            Return dictionary of a 2-tuple depending on argument
            :code:`full_output`:

            - `False`: An dictionary of simulation results keyed by state variables and each entry is of shape :code:`(num, len(t))`
            - `True`: A 2-tuple where the first entry is as above, and the second entry is the ode result from either :py:mod:`scipy.integrate.odeint` or :py:mod:`scipy.integrate.solve_ivp`. The second entry will be :code:`None` if solver is :code:`Euler`
        """
        if reset_initial_state:
            self.reset()

        t_long = t * self.Time_Scale
        d_t = t_long[1] - t_long[0]

        if verbose:
            pbar = tqdm(total=len(t_long))

        if I_ext is not None:
            if I_ext.ndim == 1:
                I_ext = np.repeat(I_ext[:, None], self.num, axis=-1)
            elif I_ext.ndim != 2:
                raise err.CompNeuroModelError("I_ext must be 1D or 2D")
            if len(I_ext) != len(t):
                raise err.CompNeuroModelError(
                    "I_ext's first dimesion must be the same length as t"
                )
            if I_ext.shape[1] > 1:
                if I_ext.shape != (len(t_long), self.num):
                    raise err.CompNeuroModelError(
                        f"I_ext expecting shape ({len(t_long)},{self.num}), "
                        f"got {I_ext.shape}"
                    )

        ode_res_info = None
        if solver.lower() == "euler":
            res = np.zeros((len(t_long), len(self._state_vec), self.num))
            res[0] = self._state_vec
            if I_ext is not None:
                for tt, (_I, _t) in enumerate(zip(I_ext[:-1], t_long[:-1])):
                    d_x = self.ode(states=self._state_vec, I_ext=_I, t=_t)
                    self._state_vec += d_t * d_x
                    self.clip()
                    res[tt + 1] = self._state_vec
                    if verbose:
                        pbar.update()
            else:
                for tt, _t in enumerate(t_long[:-1]):
                    d_x = self.ode(states=self._state_vec, t=_t)
                    self._state_vec += d_t * d_x
                    self.clip()
                    res[tt + 1] = self._state_vec
                    if verbose:
                        pbar.update()
            res = np.moveaxis(res, 0, 2)
        else:
            state_var_shape = self._state_vec.shape
            x0 = np.ravel(self._state_vec)

            # if system receives external input
            if I_ext is not None:
                interp_f = interp1d(
                    t_long, I_ext, axis=0, kind="zero", fill_value="extrapolate"
                )

                def update(t_eval, states):
                    if verbose:
                        pbar.n = int(t_eval // d_t)
                        pbar.update()
                    I_ext_t = interp_f(t_eval)
                    return self.ode(
                        states=np.reshape(states, state_var_shape),
                        I_ext=I_ext_t,
                        t=t_eval,
                    ).ravel()

            else:

                def update(t_eval, states):
                    if verbose:
                        pbar.n = int(t_eval // d_t)
                        pbar.update()
                    return self.ode(
                        states=np.reshape(states, state_var_shape), t=t_eval
                    ).ravel()

            self._ode_update = update
            if isinstance(solver, OdeSolver):
                rtol = solver_kws.pop("rtol", 1e-8)
                ode_res = solve_ivp(
                    self._ode_update,
                    t_span=(t_long.min(), t_long.max()),
                    y0=x0,
                    t_eval=t_long,
                    method=solver,
                    rtol=rtol,
                )
                ode_res_info = ode_res
                res = ode_res.y.reshape((len(self._state_vec), -1, len(t)))
            elif solver.lower() in ["lsoda", "odeint"]:
                ode_res = odeint(
                    self._ode_update,
                    y0=x0,
                    t=t_long,
                    tfirst=True,
                    full_output=full_output,
                    **solver_kws,
                )
                if full_output:
                    ode_res_y = ode_res[0]
                    ode_res_info = ode_res[1]
                    res = ode_res_y.T.reshape((len(self._state_vec), -1, len(t)))
                else:
                    res = ode_res.T.reshape((len(self._state_vec), -1, len(t)))
            else:
                rtol = solver_kws.pop("rtol", 1e-8)
                ode_res = solve_ivp(
                    self._ode_update,
                    t_span=(t_long.min(), t_long.max()),
                    y0=x0,
                    t_eval=t_long,
                    method=solver,
                    rtol=rtol,
                )
                ode_res_info = ode_res
                res = ode_res.y.reshape((len(self._state_vec), -1, len(t_long)))

        if verbose:
            pbar.close()

        res = OrderedDict({key: res[n] for n, key in enumerate(self.states.keys())})
        if full_output:
            return res, ode_res_info
        return res

    def clip(self) -> None:
        """Clip state variables based on bounds"""
        for var_name, var_val in self.states.items():
            if self.bounds[var_name] is None:
                continue
            self.states[var_name] = np.clip(
                var_val, self.bounds[var_name][0], self.bounds[var_name][1]
            )



"""Signal Generation & Analysis utilities

Adapted from the utils_ submodule of :code:`neural`

Provides the following methods:

- | :func:`generate_stimulus`: generate stimuli, currently support :code:`step`,
  | :code:`ramp` and :code:`parabolic` stimuli.
- | :func:`generate_spike_from_psth`: generate spike sequences from a PSTH.
- | :func:`compute_psth`: compute PSTH from a set of spike sequences.
- | :func:`snr`: compute Signal-to-Noise-Ratio between a signal and it\'s noisy
  | version in deciBel at each data point
- | :func:`average_snr`: compute SNR for entire signals
- | :func:`fft`: compute Fourier Transform of given signal(s), and returns the
  | spectrum as well as the frequency vector
- | :func:`nextpow2`: compute next smallest power of 2 exponent for given
  | number, same as :code:`nextpow2` in MATLAB.
- | :func:`spike_detect`: detect spike in 1D/2D variable arrays and return
  | binary array as mask for where a spike has occured.

.. _`utils`: https://github.com/chungheng/neural/blob/optimize/neural/utils.py
"""
import typing as tp
import struct
import zlib
from binascii import unhexlify
import numpy as np
from scipy.signal import butter, lfilter, find_peaks
from .. import errors as err


def chunk(btype, data):
    """Convert PNG bytes to struct"""
    return (
        struct.pack(">I", len(data))
        + btype
        + data
        + struct.pack(">I", zlib.crc32(btype + data))
    )


MINIMUM_PNG = (
    b"\x89PNG\r\n\x1A\n"
    + chunk(b"IHDR", struct.pack(">IIBBBBB", 1, 1, 8, 6, 0, 0, 0))
    + chunk(b"IDAT", unhexlify(b"789c6300010000050001"))
    + chunk(b"IEND", b"")
)


def generate_stimulus(
    mode: str,
    d_t: float,
    duration: float,
    support: tp.Tuple[float, float],
    amplitude: tp.Union[float, np.ndarray],
    sigma: float = None,
    dtype: "dtype" = np.float64,
    **kwargs,
) -> np.ndarray:
    """
    Stimuli generator

    Arguments:
        mode: shape of the waveform.
        d_t: the sampling interval for the stimuli.
        duration: the duration of the stimuli.
        support: two time points at which the stimulus starts and ends.
        amplitude: the amplitudes of the stimuli.
        sigma: variance of zero-mean Gaussian noise added to the waveform.
        ratio: a real number between 0,1 that indicates the point within
            :code:`support` where stimulus reaches the peak. Only applicable
            if :code:`mode` is either `"parabola"` or `"ramp"`.
    """

    def _generate_step(waveforms, d_t, support, amplitude, **kwargs):
        """
        Generate a set of step stimuli.

        No extra keyword argument is needed.
        """
        start = int(support[0] // d_t)
        stop = int(support[1] // d_t)

        for wav, amp in zip(waveforms, amplitude):
            wav[start:stop] = amp

    def _generate_ramp(waveforms, d_t, support, amplitude, **kwargs):
        """
        Generate a set of ramp stimuli.

        keyword arguments:
            ratio (float): a real number between 0 and 1. The point between
                `start` and `stop` where the stimulus reachs its peak.
        """
        ratio = kwargs.pop("ratio", 0.9)

        start = int(support[0] // d_t)
        stop = int(support[1] // d_t)
        peak = int((1.0 - ratio) * start + ratio * stop)

        for wav, amp in zip(waveforms, amplitude):
            wav[start:peak] = np.linspace(0.0, amp, peak - start)
            wav[peak:stop] = np.linspace(amp, 0.0, stop - peak)

    def _generate_parabola(waveforms, d_t, support, amplitude, **kwargs):
        """
        Generate a set of parabolic stimuli.

        keyword arguments:
            ratio (float): a real number between 0 and 1. The point between
                `start` and `stop` where the stimulus reachs its peak.
        """
        ratio = kwargs.pop("ratio", 0.95)

        start = int(support[0] // d_t)
        stop = int(support[1] // d_t)
        peak = int((1.0 - ratio) * start + ratio * stop)

        for wav, amp in zip(waveforms, amplitude):
            wav[start:peak] = amp * np.linspace(0.0, 1, peak - start) ** 2
            wav[peak:stop] = amp * np.linspace(1, 0.0, stop - peak) ** 2

    Nt = int((duration + d_t / 2) // d_t)

    shape = (len(amplitude), Nt) if hasattr(amplitude, "__len__") else (Nt,)
    waveforms = np.zeros(shape, dtype=dtype)

    if isinstance(mode, str):
        tmp = "_generate_%s" % mode
        assert tmp in locals(), f"Stimulus type {mode} is not supported..."
        generator = locals()[tmp]

    # ad-hoc way to deal with amplitude being a scalar or a list
    if hasattr(amplitude, "__len__"):
        generator(waveforms, d_t, support, amplitude, **kwargs)
    else:
        generator([waveforms], d_t, support, [amplitude], **kwargs)

    if sigma is not None:
        waveforms += sigma * np.random.rand(shape)
    return waveforms


def generate_spike_from_psth(
    d_t: float, psth: np.ndarray, num: int = 1, seed: int = None
) -> np.ndarray:
    """
    Generate spike sequeces from a PSTH.

    Arguments:
        d_t: the sampling interval of the input waveform.
        psth: the spike rate waveform.
        num: number of signals to generate
        seed: seed for random number generator

    Returns:
        order :code:`"C"` ndarray of type :code:`int` of shape

            1. :code:`(num, len(psth))` if :code:`num > 1`
            2. :code:`(len(psth), )` if :code:`num == 1`
    """
    psth = np.squeeze(psth)
    if psth.ndim != 1:
        raise err.CompNeuroSignalError(
            f"Only 1D PSTH is currently accepted, got {psth.ndim} instead"
        )

    rng = np.random.RandomState(seed)  # pylint:disable=E1101
    shape = (len(psth), num)
    spikes = np.zeros(shape, dtype=int, order="C")

    randvar = rng.rand(*shape)
    spikes = randvar < d_t * psth[:, None]

    return np.squeeze(spikes).T


def compute_psth(
    spikes: np.ndarray, d_t: float, window: float, interval: float
) -> tp.Tuple[np.ndarray, np.ndarray]:
    """
    Compute the peri-stimulus time histogram.

    Arguments:
        spikes: spike sequences.
        d_t: time step.
        window: the size of the window.
        interval: the time shift between two consecutive windows.

    Returns:
        A 2-tuple of

            1. the average spike rate for each windows.
            2. the time stamp for each windows.
    """

    if len(spikes.shape) > 1:
        axis = int(spikes.shape[0] > spikes.shape[1])
        spikes = np.mean(spikes, axis=axis)

    cum_spikes = np.cumsum(spikes)

    start = np.arange(0.0, d_t * len(spikes) - window, interval) // d_t
    stop = np.arange(window, d_t * len(spikes) - d_t, interval) // d_t
    start = start.astype(int, copy=False)
    stop = stop.astype(int, copy=False)

    start = start[: len(stop)]

    rates = (cum_spikes[stop] - cum_spikes[start]) / window
    stamps = np.arange(0, len(rates) * interval - d_t, interval)

    return rates, stamps


def snr(u: np.ndarray, u_rec: np.ndarray, err_bias: float = 0.0) -> np.ndarray:
    """Compute Signal to Noise Ratio at Every Data Point

    Computes the SNR according to the formula
    :math:`SNR[u, u_{rec}](t) = 10\log_{10} \\frac{u(t)^2}{err_{bias} + (u(t)-u_{rec}(t))^2}`

    Parameters:
        u: Clean Signal
        u_rec: Noisy Signal
        err_bias: bias term to be added to the denonimator of SNR term to
            avoid numerical error. Default as `0.` which results in infinite
            SNR for points where there is no error.

    Returns:
        Signal-to-Noise-Ratio in deciBel

    .. seealso:: :func:`average_snr`
    """
    _err = u - u_rec
    _snr = np.full(_err.shape, np.inf, dtype=u.dtype)
    mask = _err != 0
    _snr[mask] = 10 * np.log10(u[mask] ** 2 / _err[mask] ** 2)
    return _snr


def average_snr(u: np.ndarray, u_rec: np.ndarray, err_bias: float = 0.0) -> float:
    """Compute Avarege Signal to Noise Ratio

    Compute SNR of the entire signal instead of at every point as
    :math:`SNR[u, u_{rec}] = 10\log_{10} \\frac{E[u^2]}{err_{bias} + E[(u-u_{rec})^2]}`

    Parameters:
        u: Clean Signal
        u_rec: Noisy Signal
        err_bias: bias term to be added to the denonimator of SNR term to
            avoid numerical error. Default as `0.` which results in infinite
            SNR for points where there is no error.

    Returns:
        Scalar value Signal-to-Noise-Ratio in deciBel

    .. seealso:: :func:`snr`
    """
    _err = u - u_rec
    _err_pow = np.mean(_err ** 2)
    _sig_pow = np.mean(u ** 2)
    return 10 * np.log10(_sig_pow / (err_bias + _err_pow))


def random_signal(
    t: np.ndarray, bw: float = None, num: int = 1, seed: int = None
) -> np.ndarray:
    """Generate Random Signal

    Parameters:
        t: time points
        bw: bandwidth of output signal in Hz. If specified, lowpass filter white
            noise signal with butterworth filter. If :code:`None` (default),
            return white noise signal.
        num: number of signals to generate.
        seed: seed for random number generator.

    Returns:
        A float ndarray of shape :code:`(num, len(t))` that is of bandwidth
        :code:`bw`.
    """
    if isinstance(seed, np.random.RandomState):  # pylint:disable=E1101
        rng = seed
    else:
        rng = np.random.RandomState(seed)  # pylint:disable=E1101

    wn = rng.randn(num, len(t))
    if bw is None:
        return wn
    fs = 1 / (t[1] - t[0])
    b, a = butter(5, bw, btype="low", analog=False, fs=fs)
    sig = lfilter(b, a, wn, axis=-1)
    sig_pow = np.mean(sig ** 2, axis=-1)
    sig /= np.sqrt(sig_pow)[:, None]  # RMS power normalization
    return sig


def nextpow2(n: "Number") -> float:
    """Find Minimum Power 2 Exponent"""
    return np.ceil(np.log2(n))


def fft(
    signal: np.ndarray,
    axis: int = -1,
    extra_power2: int = None,
    fftshift: bool = True,
    fs: float = None,
    dt: float = None,
) -> tp.Tuple[np.ndarray, np.ndarray]:
    """Compute Spectrum of Signal

    Parameters:
        signal: Signal to take fft of.
        axis: axis long which to take fft, default to last
        extra_power2: extra power of 2 to add to fft when computing NFFT.
            setting it to :code:`None` will not pad the signal.
        fftshift: seeting to :code:`True` (default) will shift the signal to be
            centered at 0 Hz.
        fs: sampling frequency of the signal
        dt: temporal resolution of the signal

    .. note::

        Sampling Frequency :code:`fs` and temporal resolution :code:`dt` are
        reciprocals of each other.

    Raises:
        err.CompNeuroSignalError: Raised if :code:`fs` and :code:`dt` are both
            specified but are not reciprocals or each other.


    Returns:
        A 2-tuple of frequency (in Hz) and Complex spectrum both centered
        at 0 Hz.
    """
    if fs is None and dt is None:
        dt = 1.0
    elif fs is not None and dt is not None:
        if fs != 1.0 / dt:
            raise err.CompNeuroSignalError(
                f"Sampling Frequency fs={fs} but temporal resolution dt={dt}, "
                "they must be reciprocals of each other"
            )
    elif fs is not None and dt is None:
        dt = 1.0 / fs

    Nt = signal.shape[axis]
    if extra_power2 is None:
        nfft = signal.shape[axis]
    else:
        nfft = 2 ** int(nextpow2(Nt) + extra_power2)
    spec = np.fft.fft(signal, n=nfft, axis=axis)
    freq = np.fft.fftfreq(nfft, d=dt * (Nt / nfft))
    if fftshift:
        spec = np.fft.fftshift(spec, axes=axis)
        freq = np.fft.fftshift(freq, axes=axis)
    return freq, spec


def spike_detect(voltage: np.ndarray, **find_peaks_args) -> np.ndarray:
    """Detect Spikes from voltage trace

    Arguments:
        voltage: 1D or 2D numpy array of voltage traces.

    Keyword Arguments:
        find_peaks_args: any arguments accepted by
        :py:mod:`scipy.signal.find_peaks`

    Returns:
        A 1D or 2D binary array of same shape as input :code:`voltage`
        where :code:`True` indicates a spike

    Example:

        >>> volt = np.array([0., 1., 0., 0., 5., 2., 0.])
        >>> spk_mask = spike_detect(volt, threshol=1)
        >>> tk_idx = np.where(spk_mask)
    """
    if voltage.ndim not in [1, 2]:
        raise err.CompNeuroSignalError(
            f"Input voltage array is of dimension {voltage.ndim}, expect 1 or 2"
        )
    One_D = voltage.ndim == 1
    voltage_2d = np.atleast_2d(voltage)
    spike_mask = np.full(voltage_2d.shape, False, dtype=bool)
    for n, volt in enumerate(voltage_2d):
        tk_idx, _ = find_peaks(volt, **find_peaks_args)
        spike_mask[n, tk_idx] = True
    if One_D:
        return spike_mask[0]
    return spike_mask



#!/usr/bin/env python

"""Tests for `compneuro` package."""

import pytest
import numpy as np
from compneuro.utils.signal import average_snr
from scipy import signal


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


def assert_snr(signal1, signal2, threshold: float = 30.0):
    _asnr = average_snr(signal1, signal2)
    if _asnr >= threshold:
        return None
    else:
        raise AssertionError(
            f"SNR of {_asnr:.3f} is not greater than threshold {threshold:.3f}\n"
            f"signal1: {signal1}\n"
            f"signal2: {signal2}"
        )



# pylint:disable=C0103

from collections import OrderedDict
import numpy as np
from ..base_model import BaseModel


class HodgkinHuxley(BaseModel):
    """Hodgkin Huxley Neuron Model"""

    Time_Scale = 1e3  # s to ms
    """Hodgkin-Huxley Operates on Milisecond Scale"""
    Default_States = OrderedDict(
        V=-60, n=(0.0, 0.0, 1.0), m=(0.0, 0.0, 1.0), h=(1.0, 0.0, 1.0)
    )
    """Default State Variables of the Hodgkin-Huxley Model"""
    Default_Params = OrderedDict(
        g_Na=120.0, g_K=36.0, g_L=0.3, E_Na=50.0, E_K=-77.0, E_L=-54.387
    )
    """Default Parameters of the Hodgkin-Huxley Model"""

    def ode(self, t, states, I_ext):
        """Hodgkin-Huxley gradient function"""
        V, n, m, h = states

        alpha = np.exp(-(V + 55.0) / 10.0) - 1.0
        beta = 0.125 * np.exp(-(V + 65.0) / 80.0)
        d_n = np.zeros_like(alpha)
        _mask = np.abs(alpha) <= 1e-7
        d_n[_mask] = 0.1 * (1.0 - n[_mask]) - beta[_mask] * n[_mask]
        _mask = np.logical_not(_mask)
        d_n[_mask] = (-0.01 * (V[_mask] + 55.0) / alpha[_mask]) * (
            1.0 - n[_mask]
        ) - beta[_mask] * n[_mask]

        alpha = np.exp(-(V + 40.0) / 10.0) - 1.0
        beta = 4.0 * np.exp(-(V + 65.0) / 18.0)
        _mask = np.abs(alpha) <= 1e-7
        d_m = np.zeros_like(alpha)
        d_m[_mask] = (1.0 - m[_mask]) - beta[_mask] * m[_mask]
        _mask = np.logical_not(_mask)
        d_m[_mask] = (-0.1 * (V[_mask] + 40.0) / alpha[_mask]) * (
            1.0 - m[_mask]
        ) - beta[_mask] * m[_mask]

        alpha = 0.07 * np.exp(-(V + 65.0) / 20.0)
        beta = 1.0 / (np.exp(-(V + 35.0) / 10.0) + 1.0)
        d_h = alpha * (1 - h) - beta * h

        i_na = self.params["g_Na"] * np.power(m, 3) * h * (V - self.params["E_Na"])
        i_k = self.params["g_K"] * np.power(n, 4) * (V - self.params["E_K"])
        i_l = self.params["g_L"] * (V - self.params["E_L"])

        d_V = I_ext - i_na - i_k - i_l
        return np.vstack([d_V, d_n, d_m, d_h])



import pytest
import numpy as np
from collections import OrderedDict
from scipy.integrate import solve_ivp, cumtrapz
from compneuro.base_model import BaseModel
from compneuro.errors import CompNeuroModelError


class DummyNeuron(BaseModel):
    Default_Params = OrderedDict(a=1.0, b=0.1)
    Default_States = OrderedDict(x1=0.0, x2=(1.0, 0.0, 10.0))

    def ode(self, t, states, I_ext=None):
        x1, x2 = states
        d_x1 = -x1 * self.params["a"]
        d_x2 = I_ext
        return np.vstack([d_x1, d_x2])


def res_ref(neuron: DummyNeuron, t: np.ndarray, I_ext: np.ndarray):
    return (
        neuron.states["x1"][:, None] * np.exp(-np.outer(neuron.params["a"], t)),
        np.atleast_2d(np.cumsum(I_ext, axis=0).T * (t[1] - t[0]))
        + neuron.states["x2"][:, None],
    )


def test_neuron_init():
    neuron = DummyNeuron(a=10.0)
    assert neuron.params["a"] == 10.0
    assert neuron.states["x1"] == np.array([0.0])
    assert neuron.states["x2"] == np.array([1.0])
    assert neuron.bounds["x2"][0] == 0.0
    assert neuron.bounds["x2"][1] == 10.0
    assert neuron.bounds["x1"] is None
    np.testing.assert_equal(neuron._state_vec, np.array([[0.0], [1.0]]))

    neuron = DummyNeuron(a=10.0, num=10)
    assert neuron.params["a"] == 10.0
    np.testing.assert_equal(neuron.states["x1"], np.full((10,), 0.0))
    np.testing.assert_equal(neuron.states["x2"], np.full((10,), 1.0))
    assert neuron.bounds["x2"][0] == 0.0
    assert neuron.bounds["x2"][1] == 10.0
    assert neuron.bounds["x1"] is None
    np.testing.assert_equal(
        neuron._state_vec, np.vstack([np.full((10,), 0.0), np.full((10,), 1.0)])
    )


def test_neuron_ode():
    neuron = DummyNeuron(a=10.0, x1=10.0)
    grad = neuron.ode(t=0.0, states=neuron._state_vec, I_ext=0.0)
    np.testing.assert_equal(grad, np.vstack([-neuron.states["x1"] * 10.0, 0.0]))


def test_neuron_solve():
    dt = 1e-4
    t = np.arange(0, 0.1, dt)
    np.random.seed(0)
    I_ext = np.clip(np.random.randn(*t.shape), 0, None)
    neuron = DummyNeuron(num=1, a=1.0, x1=10.0)
    x1_ref, x2_ref = res_ref(neuron, t, I_ext)

    res = neuron.solve(t, I_ext=I_ext, solver="Euler")
    assert res["x1"].shape == (1, len(t))
    assert res["x2"].shape == (1, len(t))
    np.testing.assert_allclose(x1_ref, res["x1"], atol=1e-2, rtol=1e-2)
    np.testing.assert_allclose(x2_ref, res["x2"], atol=1e-2, rtol=1e-2)

    res2 = neuron.solve(t, I_ext=I_ext, solver="lsoda")
    assert res2["x1"].shape == (1, len(t))
    assert res2["x2"].shape == (1, len(t))
    np.testing.assert_allclose(res["x1"], res2["x1"], atol=1e-2, rtol=1e-2)
    np.testing.assert_allclose(res["x2"], res2["x2"], atol=1e-2, rtol=1e-2)

    neuron = DummyNeuron(num=10, a=10.0, x1=10.0)
    res = neuron.solve(t, I_ext=I_ext, solver="RK45", rtol=1e-5)
    x1_ref, x2_ref = res_ref(neuron, t, I_ext)
    assert res["x1"].shape == (10, len(t))
    assert res["x2"].shape == (10, len(t))
    # check to make sure all components respond similarly
    np.testing.assert_array_almost_equal(res["x1"].std(0), 0)
    np.testing.assert_array_almost_equal(res["x2"].std(0), 0)
    np.testing.assert_allclose(x1_ref, res["x1"], atol=1e-2, rtol=1e-2)
    np.testing.assert_allclose(x2_ref, res["x2"], atol=1e-2, rtol=1e-2)

    I_ext = np.clip(np.random.randn(len(t), 10), 0, None)
    res = neuron.solve(t, I_ext=I_ext, solver="lsoda")
    assert res["x1"].shape == (10, len(t))
    assert res["x2"].shape == (10, len(t))
    x1_ref, x2_ref = res_ref(neuron, t, I_ext)
    np.testing.assert_allclose(x1_ref, res["x1"], atol=1e-2, rtol=1e-2)
    np.testing.assert_allclose(x2_ref, res["x2"], atol=1e-2, rtol=1e-2)

    # check I_ext dimensionality
    num = 10
    neuron = DummyNeuron(num=num)
    # 1. If I_ext is 1D, it should work with anything
    I_ext = np.random.randn(
        len(t),
    )
    neuron.solve(t, I_ext=I_ext, solver="Euler")
    neuron.solve(t, I_ext=I_ext, solver="odeint")
    neuron.solve(t, I_ext=I_ext, solver="RK45")
    # 2. If I_ext is 2D, it must have the right shape
    I_ext = np.random.randn(len(t), num)
    neuron.solve(t, I_ext=I_ext, solver="Euler")
    neuron.solve(t, I_ext=I_ext, solver="odeint")
    neuron.solve(t, I_ext=I_ext, solver="RK45")
    # 3. If I_ext is 2D, it must have the right shape
    I_ext = np.random.randn(len(t), num + 1)
    with pytest.raises(CompNeuroModelError, match=r".* expecting shape .*"):
        neuron.solve(t, I_ext=I_ext, solver="Euler")
    with pytest.raises(CompNeuroModelError, match=r".* expecting shape .*"):
        neuron.solve(t, I_ext=I_ext, solver="odeint")
    with pytest.raises(CompNeuroModelError, match=r".* expecting shape .*"):
        neuron.solve(t, I_ext=I_ext, solver="RK45")
    # 4. If I_ext is not 1 or 2D, it will error
    I_ext = np.random.randn(len(t), 2, num + 1)
    with pytest.raises(CompNeuroModelError, match=r".* must be 1D or 2D"):
        neuron.solve(t, I_ext=I_ext, solver="Euler")



"""Comp Neuro Type Definitions

Defines custom type definitions to help with documentation and sphinx resolution
"""

import typing as tp
import scipy

solvers = tp.Union["Euler", "scipy.integrate.odeint", str, scipy.integrate.OdeSolver]



"""
Plotting functions.

Largely taken from Neural_

Provides the following methods:

- :func:`plot_spikes`: Plot 2D array of binary spike mask
- :func:`plot_mat`: Plot matrix data using :py:func:`matplotlib.pyplot.pcolormesh`
- :func:`yyaxis`: compute PSTH from a set of spike sequences.

.. _Neural: https://github.com/chungheng/neural/blob/master/neural/plot.py
"""
import typing as tp
import matplotlib.pyplot as plt
import numpy as np
from .. import errors as err


def plot_spikes(
    spikes: np.ndarray,
    dt: float = None,
    t: np.ndarray = None,
    ax: plt.Axes = None,
    markersize: int = None,
    color: tp.Union[str, tp.Any] = "k",
) -> plt.Axes:
    """
    Plot Spikes in raster format

    Arguments:
        spikes: the spike states in binary format, where 1 stands for a spike.
            The shape of the spikes should either be (N_times, ) or (N_trials, N_times)
        dt: time resolution of the time axis.
        t: time axes for the spikes, use arange if not provided
        ax: which axis to plot into, create one if not provided
        markersize: size of raster
        color: color for the raster. Any acceptable type of `matplotlib.pyplot.plot`'s
            color argument is accepted.

    .. note::

            If `t` is specified, it is assumed to have the same
            length as `mat.shape[1]`, which is used to find the x coordinate of
            the spiking values of the data. If `t` is
            not specified, the time-axis is formated by resolution `dt`.
            `dt` is assumed to be 1 if not specified.

    Returns:
        The axis that the raster is plotted into
    """
    spikes = np.atleast_2d(spikes)
    if spikes.ndim != 2:
        raise err.CompNeuroPlotError(
            f"matrix need to be of ndim 2, (channels x time), got ndim={spikes.ndim}"
        )

    if t is not None:
        if len(t) != spikes.shape[1]:
            raise err.CompNeuroPlotError(
                "Time vector 't' does not have the same shape as the matrix."
                f" Expected length {spikes.shape[1]} but got {len(t)}"
            )
    else:
        if dt is None:
            dt = 1.0
        else:
            if not np.isscalar(dt):
                raise err.CompNeuroPlotError("dt must be a scalar value.")
        t = np.arange(spikes.shape[1]) * dt

    if ax is None:
        fig = plt.gcf()
        ax = fig.gca()

    neu_idx, t_idx = np.nonzero(spikes)

    try:
        ax.plot(t[t_idx], neu_idx, "|", c=color, markersize=markersize)
    except ValueError as e:
        raise err.CompNeuroPlotError(
            "Raster plot failed, likely an issue with color or markersize setting"
        ) from e
    except IndexError as e:
        raise err.CompNeuroPlotError(
            "Raster plot failed, likely an issue with spikes and time vector mismatch"
        ) from e
    except Exception as e:
        raise err.CompNeuroPlotError("Raster plot failed due to unknown error") from e
    ax.set_xlim([t.min(), t.max()])
    return ax


def plot_mat(
    mat: np.ndarray,
    dt: float = None,
    dy: float = None,
    t: np.ndarray = None,
    y: np.ndarray = None,
    ax: plt.Axes = None,
    cax=None,
    vmin: float = None,
    vmax: float = None,
    cbar_kw: dict = None,
    cmap: tp.Any = None,
    shading: str = "nearest",
) -> tp.Union[tp.Tuple[plt.Axes, tp.Any], plt.Axes]:
    """
    Plot Matrix with formatted time axes

    Arguments:
        mat: the matrix to be plotted, it should of shape (N, Time)
        dt: time resolution of the time axis.
        dy: resolution of the Y-axis
        t: time axes for the matrix, use arange if not provided.
        y: spatial axes of the matrix, use arange if not provided.


        ax: which axis to plot into, create one if not provided
        cax: which axis to plot colorbar into
            - if instance of axis, plot into that axis
            - if is True, steal axis from `ax`
        vmin: minimum value for the imshow
        vmax: maximum value for the imshow
        cbar_kw: keyword arguments to be passed into the colorbar creation
        cmap: colormap to use
        shading: optional, argument to be passed to `matplotlib.pyplot.pcolormesh`

    .. note::

        If `t` is specified, it is assumed to have the same
        length as `mat.shape[1]`. Consequently, the x-axis will be formatted
        to take the corresponding values from `t` based on index. If `t` is
        not specified, the time-axis is formated by resolution `dt`.
        If neither are specified, `dt` is assumed to be 1.


    .. note::

        If `y` is specified, it is assumed to have the same
        length as `mat.shape[0]`. Consequently, the y-axis will be formatted
        to take the corresponding values from `y` based on index. If `y` is
        not specified, the time-axis is formated by resolution `dy`.
        If neither are specified, `dy` is assumed to be 1.

    Returns:
        A 2-tuple of

            1. the axis that the raster is plotted into
            2. colorbar object, only returned if cax is `True` or a `plt.Axes` instance
    """
    mat = np.atleast_2d(mat)
    if mat.ndim != 2:
        raise err.CompNeuroPlotError(
            "matrix need to be of ndim 1 (N_time),or ndim 2 (N_trials x N_times),"
            f" got ndim={mat.ndim}"
        )
    if t is not None:
        if len(t) != mat.shape[1]:
            raise err.CompNeuroPlotError(
                "Time vector 't' does not have the same shape as the matrix."
                f" Expected length {mat.shape[1]} but got {len(t)}"
            )
    else:
        if dt is None:
            dt = 1
        t = np.arange(mat.shape[1]) * dt

    if y is not None:
        if len(y) != mat.shape[0]:
            raise err.CompNeuroPlotError(
                "Spatial vector 'y' does not have the same shape as the matrix."
                f" Expected length {mat.shape[0]} but got {len(y)}"
            )
    else:
        if dy is None:
            dy = 1
        y = np.arange(mat.shape[0]) * dy

    if ax is None:
        fig = plt.gcf()
        ax = fig.gca()

    cim = ax.pcolormesh(t, y, mat, vmin=vmin, vmax=vmax, cmap=cmap, shading=shading)

    if cax:
        if cbar_kw is None:
            cbar_kw = {}
        if not isinstance(cax, plt.Axes):
            cbar = plt.colorbar(cim, ax=ax, **cbar_kw)
        else:
            cbar = plt.colorbar(cim, cax, **cbar_kw)
        return ax, cbar
    return (ax,)


def yyaxis(ax: plt.Axes, c: tp.Any = "red") -> plt.Axes:
    """
    Create a second colored axis

    Arguments:
        ax: the main axis to generate a twinx from
        c: color of the twinx, see matplotlib\'s colors_

    Returns:
        The new axis object

    .. note::

        This method will only make the twinx look like the color in
        MATLAB's :code:`yyaxis` function. However, unlike in MATLAB,
        it will not set the linestyle and linecolor of the lines that
        are plotted after twinx creation.

    .. _colors: https://matplotlib.org/stable/gallery/color/color_demo.html
    """
    ax2 = ax.twinx()
    ax2.spines["right"].set_color(c)
    ax2.tick_params(axis="y", colors=c, which="both")
    ax2.yaxis.label.set_color(c)
    return ax2



import pytest
from scipy.signal import find_peaks
import numpy as np
from collections import OrderedDict
from compneuro.base_model import BaseModel
from compneuro.neurons.hodgkin_huxley import HodgkinHuxley
from compneuro.neurons.hodgkin_huxley_rinzel import HodgkinHuxleyRinzel
from compneuro.neurons.hodgkin_huxley_wilson import HodgkinHuxleyWilson
from compneuro.utils.signal import generate_stimulus
from .test_compneuro import assert_snr
from .reference_neuron_implementations import hh, hhr, hhw


@pytest.fixture
def inputs():
    dt = 1e-5
    t = np.arange(0.0, 0.1, dt)
    I_ext = generate_stimulus("step", dt, 0.1, (0.02, 0.1), 50.0)
    return dt, t, I_ext


def test_hhn(inputs):
    dt, t, I_ext = inputs
    hhn_ref_V, _ = hh(t, I_ext)
    hhn_ref_V -= 65.0
    neu = HodgkinHuxley(num=1, V=-75.0)
    res_euler = neu.solve(t, I_ext, solver="Euler", verbose=False)
    res_odeint = neu.solve(t, I_ext, solver="odeint", rtol=1e-10, verbose=False)
    res_rk45 = neu.solve(t, I_ext, solver="RK45", rtol=1e-10, verbose=False)

    assert_snr(hhn_ref_V, res_euler["V"][0], threshold=6.0)
    assert_snr(res_euler["V"], res_odeint["V"], threshold=10.0)
    assert_snr(res_euler["V"], res_rk45["V"], threshold=10.0)


def test_rinzel(inputs):
    dt, t, I_ext = inputs
    hhr_ref_V, _ = hhr(t, I_ext)
    hhr_ref_V -= 65.0
    neu = HodgkinHuxleyRinzel(num=1)
    res_euler = neu.solve(t, I_ext, solver="Euler", verbose=False)
    res_odeint = neu.solve(t, I_ext, solver="odeint", rtol=1e-10, verbose=False)
    res_rk45 = neu.solve(t, I_ext, solver="RK45", rtol=1e-10, verbose=False)

    assert_snr(hhr_ref_V, res_euler["V"][0], threshold=6.0)
    assert_snr(res_euler["V"], res_odeint["V"], threshold=10.0)
    assert_snr(res_euler["V"], res_rk45["V"], threshold=10.0)


def test_wilson(inputs):
    dt, t, I_ext = inputs
    hhw_ref_V, _ = hhw(t, I_ext)
    neu = HodgkinHuxleyWilson(num=1)
    res_euler = neu.solve(t, I_ext, solver="Euler", verbose=False)
    res_odeint = neu.solve(t, I_ext, solver="odeint", rtol=1e-10, verbose=False)
    res_rk45 = neu.solve(t, I_ext, solver="RK45", rtol=1e-10, verbose=False)

    assert_snr(hhw_ref_V, res_euler["V"][0], threshold=6.0)
    assert_snr(res_euler["V"], res_odeint["V"], threshold=10.0)
    assert_snr(res_euler["V"], res_rk45["V"], threshold=10.0)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["numpy>=1.19", "matplotlib>=3.3", "scipy>=1.7"]
dev_requirements = ["pylint", "black", "Sphinx", "sphinx-autobuild"]
dist_requirements = ["twine", "coverage"]
test_requirements = ["pytest>=3", "tox>=3.23"]

setup(
    author="Bionet Lab",
    author_email="",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    description="Python Module for Book 4020",
    install_requires=requirements,
    license="BSD license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="compneuro",
    name="compneuro",
    packages=find_packages(include=["compneuro", "compneuro.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    extras_require={
        "dev": dev_requirements,
        "dist": dev_requirements + dist_requirements,
    },
    url="https://github.com/TK-21st/CompNeuro",
    version="0.1.1",
    zip_safe=False,
)



# pylint:disable=C0103
from collections import OrderedDict
import numpy as np
from ..base_model import BaseModel


class HodgkinHuxleyWilson(BaseModel):
    """Wilson Neuron Model

    The Wilson Neuron Model is a reduced version of the Hodgkin-Huxley Neuron
    model. Instead of having 4 state variables :math:`(V, n, m, h)`, wilson
    only has 2 state variables :math:`(V, R)`.
    """

    Time_Scale = 1e3
    """Wilson Operates on Milisecond Scale"""
    Default_States = OrderedDict(V=-70, R=0.088)
    """Default State Variables of the Wilson Model"""
    Default_Params = OrderedDict(C=1.2, E_K=-92.0, g_K=26.0, E_Na=55.0)
    """Default Parameters of the Wilson Model"""

    def ode(self, t, states, I_ext):
        """Wilson gradient function"""
        V, R = states

        R_infty = 0.0135 * V + 1.03

        d_R = (R_infty - R) / 1.9

        I_Na = (17.81 + 0.4771 * V + 0.003263 * V ** 2) * (V - self.params["E_Na"])
        I_K = self.params["g_K"] * R * (V - self.params["E_K"])
        d_V = 1.0 / self.params["C"] * (I_ext - I_Na - I_K)

        return np.vstack([d_V, d_R])



# pylint:disable=C0103

from collections import OrderedDict
import numpy as np
from ..base_model import BaseModel


class HodgkinHuxleyRinzel(BaseModel):
    """Rinzel Neuron Model

    The Rinzel Neuron Model is a reduced version of the Hodgkin-Huxley Neuron
    model. Instead of having 4 state variables :math:`(V, n, m, h)`, rinzel
    only has 2 state variables :math:`(V, W)`.

    .. note::

        In some texts, the state variables :math:`W` may also be called
        :math:`R`.
    """

    Time_Scale = 1e3  # s to ms
    """Rinzel Operates on Milisecond Scale"""

    Default_States = OrderedDict(V=-65, W=(0.0, 0.0, 1.0))
    """Default State Variables of the Rinzel Model"""

    Default_Params = OrderedDict(
        C=1.0,
        g_Na=120.0,
        g_K=36.0,
        g_L=0.3,
        E_Na=50.0,
        E_K=-77.0,
        E_L=-54.387,
        S=1.27135220916422,
    )
    """Default Parameters of the Rinzel Model"""

    def ode(self, t, states, I_ext):
        """Rinzel gradient function"""
        V, W = states

        alpha_n = 0.01 * (10 - (V + 65)) / (np.exp((10 - (V + 65)) / 10) - 1)
        alpha_m = 0.1 * (25 - (V + 65)) / (np.exp((25 - (V + 65)) / 10) - 1)
        alpha_h = 0.07 * np.exp(-(V + 65) / 20)

        beta_n = 0.125 * np.exp(-(V + 65) / 80)
        beta_m = 4 * np.exp(-(V + 65) / 18)
        beta_h = 1 / (np.exp((30 - (V + 65)) / 10) + 1)

        n_infty = alpha_n / (alpha_n + beta_n)
        m_infty = alpha_m / (alpha_m + beta_m)
        h_infty = alpha_h / (alpha_h + beta_h)
        W_infty = (
            self.params["S"]
            / (1 + self.params["S"] ** 2)
            * (n_infty + self.params["S"] * (1 - h_infty))
        )

        tau_W = 1 + 5 * np.exp(-(((V + 65) - 10) ** 2) / 55 ** 2)

        d_W = 3.0 * W_infty / tau_W - 3.0 * W / tau_W

        # Update the ionic currents and membrane voltage:
        I_K = (
            self.params["g_K"] * (W / self.params["S"]) ** 4 * (V - self.params["E_K"])
        )
        I_Na = self.params["g_Na"] * m_infty ** 3 * (1 - W) * (V - self.params["E_Na"])
        I_L = self.params["g_L"] * (V - self.params["E_L"])
        d_V = (I_ext - I_K - I_Na - I_L) / self.params["C"]

        return np.vstack([d_V, d_W])
